from typing import Sequence
import asyncio
import time
import uuid

from chai_py.chai_bot import ChaiBot
from chai_py.types import Update, LatestMessage, Message, MessageKind


class TRoom():
    """Code for testing your bot locally. """

    def __init__(self, bots: Sequence[ChaiBot]):
        self.bots = bots
        self.conversation_id = str(uuid.uuid4())
        self.messages = []
        if len(self.bots) == 0:
            raise RuntimeError("Cannot start with no bots.")
        print("Starting TRoom. Press ctrl-c to escape.")
        self.setup_bots()

    def chat(self):
        asyncio.run(self._loop())

    def test_chat(self, messages):
        asyncio.run(self._message_loop(messages))

    async def _message_loop(self, messages):
        await self.send_message(self.bots[0].FIRST_MESSAGE_STRING)

        for message in messages:
            await self.send_message(message)

    async def _loop(self):
        await self.send_message(self.bots[0].FIRST_MESSAGE_STRING)
        while True:
            try:
                message = input("Enter your message: ")
            except KeyboardInterrupt:
                print("Interrupted TRoom.")
                break
            await self.send_message(message)

    def setup_bots(self):
        async def get_messages_local(conversation_id: str):
            return self.messages

        for bot in self.bots:
            bot.get_messages = get_messages_local

    async def send_message(self, message: str):
        timestamp = int(time.time() * 1000)
        print(f"<<< {message}")
        self.messages.append(
            Message(
                sender_uid='__local_dev',
                timestamp=timestamp,
                message_kind=MessageKind.TEXT,
                content=message,
            )
        )
        update = Update(
            conversation_id=self.conversation_id,
            latest_message=LatestMessage(
                text=message,
                timestamp=timestamp
            )
        )

        for coro in asyncio.as_completed([self._bot_on_message(bot, update) for bot in self.bots]):
            bot, result = await coro
            print(f">>> {result}")
            self.messages.append(
                Message(
                    sender_uid=bot._bot_uid,
                    timestamp=timestamp,
                    message_kind=MessageKind.TEXT,
                    content=result,
                )
            )

    @staticmethod
    async def _bot_on_message(bot: ChaiBot, update: Update):
        result = await bot.on_message(update)
        return bot, result


TestRoom = TRoom



from chai_py.defaults import GUEST_UID, GUEST_KEY
from chai_py.logger import logger


def get_auth():
    return auth


def set_auth(uid: str, key: str):
    """
    Sets package-wide developer authentication.

    :param uid: Developer Unique Identifier.
    :param key: Developer key.
    :return:
    """
    global auth
    auth = ChaiAuth(uid=uid, key=key)


class ChaiAuth():
    """Credentials for developer API. """

    def __init__(self, uid: str, key: str):
        self._uid = uid
        self._key = key
        self._accessed = False

    @property
    def is_guest(self):
        return self._uid == GUEST_UID

    @property
    def uid(self) -> str:
        if not self._accessed and self.is_guest:
            logger.warning("Using guest credentials. To be able to redeploy and monitor your Chat AIs, "
                           "run chai_py.set_auth() with your uid and key from https://chai.ml/dev/")
            self._accessed = True
        return self._uid

    @property
    def key(self) -> str:
        return self._key

    def __repr__(self):
        return f"{self.__class__.__name__}(uid={self._uid}, key={self._key})"


auth = ChaiAuth(uid=GUEST_UID, key=GUEST_KEY)



import builtins
import functools

from segno import QRCode


def check_is_notebook() -> bool:
    """Checks if the current environment is a Python notebook."""
    return hasattr(builtins, "__IPYTHON__")


def only_notebook(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if IS_NOTEBOOK:
            return func(*args, **kwargs)
        else:
            raise RuntimeError(
                "Attempting to run notebook utility in a non-notebook environment. "
                "If you are not in a notebook environment, this is a bug - let us know how you ran into it!"
            )

    return wrapper


@only_notebook
def show_qr(qr: QRCode):
    """Generates and displays a QR code."""
    from IPython.display import display, SVG
    return display(SVG(
        qr.svg_inline(scale=10, light="#fff")
    ))


@only_notebook
def register_write_and_run():
    from IPython.core.magic import register_cell_magic

    @register_cell_magic
    def write_and_run(line, cell):
        """Saves the contents of the cell to a file and reloads the class.

        Call with
            %%write_and_run directory file_name class_name
        E.g.
          %%write_and_run bot bot.py Bot
              will save the cell to bot/bot.py, and import the class Bot
        """
        from importlib import reload
        from pathlib import Path

        directory, file_name, class_name = line.split()
        file_path = Path(directory) / file_name
        file_path.parent.mkdir(exist_ok=True)

        print(f"Saving cell to {file_path} and reloading class {class_name}.")

        # Save cell to file
        with file_path.open('w') as f:
            f.write(cell)

        # Import and reload the bot class.
        exec(f"import {directory}.{file_path.stem}")
        eval(f"reload({directory}.{file_path.stem})")
        ipython = get_ipython()
        ipython.push({class_name: eval(f"{directory}.{file_path.stem}.{class_name}")})


IS_NOTEBOOK = check_is_notebook()


@only_notebook
def setup_notebook():
    try:
        import nest_asyncio
        nest_asyncio.apply()
        print("Set up nest_asyncio for TRoom.")
    except Exception:
        raise RuntimeError(
            "Error setting up nest_asyncio. If you are in a notebook, use `pip install chaipy[notebook]`. "
            "If not, this is a bug - let us know how you ran into it!"
        )

    try:
        register_write_and_run()
        print("Set up write_and_run magic.")
    except Exception:
        raise RuntimeError(
            "Error setting up write_and_run magic."
        )



DEFAULT_SIGNED_URL_CREATOR = "https://europe-west2-chai-959f8.cloudfunctions.net/signed-url-creator"
GUEST_UID = "20c900f0-e6f5-440c-bd83-e1ad894ad598"
GUEST_KEY = "LyRgaRjLmldLltvIL5Q5QY6XxNgIAMiTxxQdq2GyiJA8jsd55q7FxYcOgZUbV2dFmeK75pt4vd3zXe9Acd7pyw"

BOT_DELETER_ENDPOINT = "https://europe-west2-chai-959f8.cloudfunctions.net/bot-deleter"
DEFAULT_BOT_STATUS_ENDPOINT = "https://europe-west2-chai-959f8.cloudfunctions.net/bot-status"
DEFAULT_BOT_LOGS_ENDPOINT = "https://europe-west2-chai-959f8.cloudfunctions.net/bot-logs"
ENDPOINT = "https://europe-west2-chai-959f8.cloudfunctions.net/chai-api"

API_HOST = 'http://api.chai.ml'



from chai_py.deployment import upload_and_deploy, wait_for_deployment, get_bot_status, share_bot
from chai_py.auth import set_auth, get_auth
from chai_py.chai_bot import ChaiBot
from chai_py.cloud_logs import display_logs, get_logs, Log
from chai_py.notebook_utils import IS_NOTEBOOK, setup_notebook
from chai_py.packaging import package, Metadata
from chai_py.t_room import TRoom
from chai_py.types import LatestMessage, Update, Message, MessageKind



from dataclasses import dataclass
from enum import Enum

import requests

from chai_py.auth import get_auth
from chai_py import defaults, error


class BotStatus(Enum):
    """
    Active bots can be discovered by users on the app
    and inactive bots can only be accessed with a direct link.
    """
    ACTIVE = 1
    INACTIVE = 2


@dataclass
class DeployedBot:
    """
    A currently deployed chatbot.

    Args:
        bot_uid (str): The unique ID of the chatbot.
        name (str): The name of the chatbot.
        developer_uid (str): ID of the developer who created the chatbot.
        status (BotStatus): Whether the chatbot is visible to users.
    """
    bot_uid: str
    name: str
    developer_uid: str
    status: BotStatus


def get_bots():
    """
    Retrive a summary of all bots deployed by a developer.

    Returns:
        list[DeployedBot]: list of bots you have deployed.
    """
    url = '{}/chatbots'.format(defaults.API_HOST)
    js = {'developer_uid': _get_developer_uid()}
    resp = requests.get(url, json=js, auth=_credentials())
    _check_response_for_error(resp)
    return _parse_multiple_bots_response(resp)


def activate_bot(bot_uid: str):
    """
    Activate a bot so it can be discovered by users on the app.

    Args:
        bot_uid (str): the unique ID of the bot to make visibile
    """
    url = '{}/chatbots/{}'.format(defaults.API_HOST, bot_uid)
    js = {'status': 'active'}
    resp = requests.post(url, json=js, auth=_credentials())
    _check_response_for_error(resp)


def deactivate_bot(bot_uid: str):
    """
    Deactivate a bot so it cannot be discovered by users on the app.

    Args:
        bot_uid (str): the unique ID of the bot to remove from visibility.
    """
    url = '{}/chatbots/{}'.format(defaults.API_HOST, bot_uid)
    js = {'status': 'inactive'}
    resp = requests.post(url, json=js, auth=_credentials())
    _check_response_for_error(resp)


def _check_response_for_error(resp):
    if resp.status_code != 200:
        raise error.APIError('Bad response ({}): {}'.format(resp.status_code, resp.text))


def _get_developer_uid():
    return get_auth().uid


def _credentials():
    auth = get_auth()
    return requests.auth.HTTPBasicAuth(auth.uid, auth.key)


def _parse_multiple_bots_response(resp):
    bots = []
    for raw_bot_response in resp.json()['data']:
        bots.append(_parse_raw_bot_dict(raw_bot_response))
    return bots


def _parse_raw_bot_dict(data):
    status = BotStatus[data['status'].upper()]
    return DeployedBot(data['bot_uid'], data['name'], data['developer_uid'], status)



from datetime import datetime
from typing import List
import sys

from typing_extensions import TypedDict
import requests

from chai_py.auth import get_auth
from chai_py.defaults import ENDPOINT
from chai_py.logger import logger


class Log(TypedDict):
    timestamp: int
    severity: str
    log: str


def get_logs(bot_uid: str, errors: bool = False) -> List[Log]:
    """Retrieves logs for specified bot.

    Logs can only be pulled by the bot's developer.

    :param bot_uid: Bot UID
    :param errors: If True, only retrieves logs with severity: Error
    :return:
    """
    auth = get_auth()
    r = requests.get(
        ENDPOINT,
        params={
            'uid': auth.uid,
            'key': auth.key,
            'bot_uid': bot_uid,
            'item': 'logs',
            'errors': True if errors else None
        }
    )
    r.raise_for_status()
    return r.json()['data']


def display_logs(logs: List[Log]):
    """Prints logs with clear separation.

    :param logs: Logs retrieved with get_logs()
    :return:
    """
    for log in logs:
        timestamp = datetime.fromtimestamp(log['timestamp'] / 1000).isoformat(timespec='seconds')
        severity = log['severity']
        is_error = severity == 'ERROR'
        _log = logger.error if is_error else logger.info
        _log(f"===== LOG AT {timestamp} (SEVERITY: {severity}) =====")
        print(log['log'], file=sys.stderr if is_error else sys.stdout, flush=True)
        _log("===== END LOG =====")



from dataclasses import dataclass, field
from pathlib import Path
from typing import Type, AnyStr, List, Optional
from zipfile import ZipFile, ZIP_DEFLATED
import imghdr
import inspect
import json
import os
import pprint
import re
import shutil
import stat
import tempfile
import warnings

import pkg_resources
import requests

from chai_py.auth import get_auth
from chai_py.chai_bot import ChaiBot

MAX_SUPPORTED_MEMORY = 4096


@dataclass
class Metadata:
    """Information required for bot deployment."""
    # Name of the bot.
    name: str
    # Profile image for the bot. Has to be a valid URL.
    image_url: str
    # The alphanumeric part of a hex color code. (E.g. ffffff)
    color: str
    # Description of the bot.
    description: str
    # Python class (N.B. not object!) that inherits from ChaiBot.
    input_class: Type[ChaiBot]

    # Developer Unique ID.
    developer_uid: str = field(default_factory=lambda: get_auth().uid)

    # Total available memory for the bot in MB. This includes memory needed to store sources and data.
    memory: int = 256

    def verify(self, bot_file: Path):
        """Performs basic checks to ensure validity of the metadata."""
        assert isinstance(self.name, str)
        assert len(self.name) >= 3, "Bot name has to be at least 3 characters."

        assert len(self.description) > 0, "Bot has to have description."
        assert self.input_class.__init__ is ChaiBot.__init__, \
            "Do not override ChaiBot.__init__(). Override the setup() method instead."

        assert not (bot_file.parent / "main.py").exists(), "Do not create a main.py file in your bot's root directory."

        try:
            verify_image_url(self.image_url)
        except Exception:
            raise ValueError(f"Could not verify image url ({self.image_url})")

        assert isinstance(self.color, str)
        assert re.search(r"^(?:[0-9a-fA-F]{3}){1,2}$", self.color), \
            f"Color has to be provided as the alphanumeric part of the hex code (e.g. ffffff), found {self.color}"

        assert isinstance(self.memory, int), f"Attribute .memory has to be an integer (found type {type(self.memory)})."
        assert self.memory <= MAX_SUPPORTED_MEMORY, f"Attribute .memory has to be less than or equal to {MAX_SUPPORTED_MEMORY} (found {self.memory})."


def package(metadata: Metadata, requirements: Optional[List[str]] = None, path: Optional[str] = None):
    """Packages the chatbot into a single archive for deployment.

    Performs some preliminary checks on the metadata.
    Creates a _package.zip file in the directory containing the file that contains the bot class
    unless a path is provided.

    :param metadata:
    :param requirements:
    :param path:
    :return:
    """
    bot_file = Path(inspect.getfile(metadata.input_class))

    print("Running verification checks on metadata.")
    metadata.verify(bot_file)

    metadata_dict = {
        'name': metadata.name,
        'imageUrl': metadata.image_url,
        'color': metadata.color,
        'developerUid': metadata.developer_uid,
        'description': metadata.description,
        'inputFile': bot_file.stem,
        'inputClass': metadata.input_class.__name__,
        'memory': metadata.memory,
    }
    print("Prepared metadata:")
    pprint.pprint(metadata_dict)

    print("Preparing temporary directory...")
    with tempfile.TemporaryDirectory() as temp_dir:
        # Copy files in bot directory
        def ignore(src, names):
            ignore_list = []
            for name in names:
                # e.g .git folder is not wanted
                if name.startswith('.') or name.startswith('_package.zip'):
                    warnings.warn(
                        f"Ignoring files which start with '.': {name}.",
                        RuntimeWarning
                    )
                    ignore_list.append(name)
                if name == "main.py":
                    raise RuntimeError("Bot root directory cannot contain a main.py file.")
            return ignore_list

        copytree(bot_file.parent, temp_dir, ignore=ignore)
        # Write metadata.json
        with (Path(temp_dir) / "metadata.json").open("w") as f:
            json.dump(metadata_dict, f)

        # Write requirements.txt
        if requirements:
            write_valid_requirements_file(Path(temp_dir) / "requirements.txt", requirements)

        # Create zip
        if path is None:
            path = bot_file.parent / "_package.zip"
        else:
            path = Path(path)

        with path.open("wb") as f:
            zipfile_from_folder(temp_dir, f)
        print(f"Created zip package at {path}.")


def verify_image_url(url: str):
    """Verifies that the provided url resolves to an image.

    Performs a GET request on the given url and performs a trivial (non-conclusive) check
    that the image type can be inferred from the received bytes.

    :param url:
    :return:
    """
    r = requests.get(url)
    try:
        imghdr.what(None, h=r.content)
    except Exception:
        raise ValueError(
            f"Could not verify image type from bytes "
            f"(response content-type of {r.headers.get('content-type')})"
        )


def zipfile_from_folder(folder: AnyStr, file):
    # Adapted from https://stackoverflow.com/a/17080988
    with ZipFile(file, "w", compression=ZIP_DEFLATED) as zip_archive:
        for root, dirs, files in os.walk(folder):
            # add directory (needed for empty dirs)
            zip_archive.write(root, os.path.relpath(root, folder))
            for file in files:
                filename = os.path.join(root, file)
                if os.path.isfile(filename):  # regular files only
                    arcname = os.path.join(os.path.relpath(root, folder), file)
                    zip_archive.write(filename, arcname)


def copytree(src, dst, symlinks=False, ignore=None):
    """Copies files from src to dst.

    Taken from https://stackoverflow.com/a/22331852.
    Necessitated by Python 3.7 environment; Python 3.8's shutil.copytree can be used directly
    as it has the necessary dirs_exist_ok argument.

    :param src: Source directory
    :param dst: Target directory
    :param symlinks:
    :param ignore: Callable
    :return:
    """
    if not os.path.exists(dst):
        os.makedirs(dst)
        shutil.copystat(src, dst)
    lst = os.listdir(src)
    if ignore:
        excl = ignore(src, lst)
        lst = [x for x in lst if x not in excl]
    for item in lst:
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if symlinks and os.path.islink(s):
            if os.path.lexists(d):
                os.remove(d)
            os.symlink(os.readlink(s), d)
            try:
                st = os.lstat(s)
                mode = stat.S_IMODE(st.st_mode)
                os.lchmod(d, mode)
            except Exception:
                pass  # lchmod not available
        elif os.path.isdir(s):
            copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)


def write_valid_requirements_file(path: AnyStr, requirements: List[str]):
    """Writes a valid requirements.txt file.

    Iterates through list of requirements, writing valid entries to the specified file,
    Ignores (and prints) invalid requirements.

    :param path:
    :param requirements:
    :return:
    """
    with Path(path).open("w") as f:
        for requirement in requirements:
            try:
                pkg_resources.Requirement.parse(requirement)
                f.write(requirement + "\n")
            except Exception as e:
                print(f"Ignoring requirement {requirement}: {e}")



from dataclasses import dataclass
from enum import auto, Enum


@dataclass
class LatestMessage:
    text: str
    timestamp: int  # Milliseconds since epoch


@dataclass
class Update:
    conversation_id: str
    latest_message: LatestMessage


class MessageKind(Enum):
    TEXT = auto()
    IMAGE = auto()


@dataclass
class Message:
    sender_uid: str
    timestamp: int
    message_kind: MessageKind
    content: str



from pathlib import Path
from typing import AnyStr, Optional
import time

from halo import Halo
from requests import HTTPError
from typing_extensions import TypedDict
import requests
import segno

from chai_py.auth import get_auth
from chai_py.cloud_logs import display_logs, get_logs
from chai_py.defaults import ENDPOINT
from chai_py.notebook_utils import IS_NOTEBOOK, show_qr


def upload_and_deploy(package: AnyStr, bot_uid: str = None) -> str:
    """Uploads the given archive, triggering deployment of the chatbot.

    :param package: Path to the packaged chatbot zip.
    :param bot_uid: Used to modify an existing bot: the UID of the previously-deployed bot.
    :return bot_uid: The UID of the deployed bot.
    """
    package = Path(package)

    deployment_file = package.parent / "_deployment"

    if deployment_file.exists() and bot_uid is None:
        with deployment_file.open("r") as f:
            previous_bot_uid = f.read().strip()
        print("Detected previous deployment from this location. Use the same bot UID as before?")
        print(f" [y] (default) Yes. Update the bot ({previous_bot_uid}).")
        print(f" [n] No. Deploy as a new bot.")
        input_key = input().lower()
        if input_key == "y" or input_key == "":
            bot_uid = previous_bot_uid
            print(f"Using previous bot UID: {bot_uid}")
        elif input_key == "n":
            pass
        else:
            raise RuntimeError("Unknown input.")

    auth = get_auth()
    try:
        r = requests.post(
            ENDPOINT,
            params={'uid': auth.uid, 'key': auth.key, 'bot_uid': bot_uid}
        )
        try:
            r.raise_for_status()
        except Exception:
            raise RuntimeError(r.json())
    except Exception:
        print(r.content, r.reason)
        raise RuntimeError("Failed to retrieve signed URL.")

    url = r.text
    if bot_uid is None:
        print("Creating new bot.")
        bot_uid = parse_signed_url_for_bot_uid(url)
        print(f"Received bot UID: {bot_uid}")
    with deployment_file.open("w") as f:
        f.write(bot_uid)

    with package.open("rb") as f:
        r = requests.put(url, data=f, headers={'content-type': 'application/zip'})
        r.raise_for_status()
    print(f"Successfully uploaded {package}.")
    return bot_uid


def parse_signed_url_for_bot_uid(url: str):
    """Parses a bot UID from a signed URL.

    Assumes the signed URL follows the following scheme:
        [...]/{bot_uid}.zip?[...]

    :param url:
    :return:
    """
    endpoint = url.split("?", maxsplit=1)[0]
    file_name = endpoint.split("/")[-1]
    bot_uid = file_name.split(".")[0]
    return bot_uid


def wait_for_deployment(bot_uid: str, sleep: float = 3):
    """Waits for deployment of the bot to complete.

    :param bot_uid:
    :param sleep: Polling interval in seconds.
    :return:
    """
    MIN_SLEEP = 1
    sleep = max(MIN_SLEEP, sleep)

    BOT_DEPLOYMENT_PROCESS = [
        "signed_url_created",
        "processing_upload",
        "deploying",
        "initialized",
    ]
    current_deployment_process = -1  # Index for BOT_DEPLOYMENT_PROCESS
    completed_processes = []
    existing_status = None
    try:
        _existing_status = get_bot_status(bot_uid)
        if "activeDeployment" in _existing_status:
            existing_status = _existing_status
            print(f"Found previous deployment: Version {existing_status['activeDeployment']['version']}")
        elif "failedDeployment" in _existing_status:
            existing_status = _existing_status
            print(f"Found previous failed deployment: Version {existing_status['failedDeployment']['version']}")

    except HTTPError as e:
        if e.response.status_code != 404:
            raise e
        else:
            print("Did not find previous deployment.")

    MAXIMUM_ERROR_RETRIES = 10
    error_retries = 0

    start_time = time.time()
    with Halo(text="Polling for progress...") as spinner:
        spinner.start()
        time.sleep(5)  # Initial wait to avoid bot status error.
        while True:
            try:
                status = get_bot_status(bot_uid)
            except Exception as e:
                spinner.warn(f"Error getting bot status (UID: {bot_uid}): {e}")
                error_retries += 1
                if error_retries > MAXIMUM_ERROR_RETRIES:
                    spinner.fail()
                    print(f"Hit retry-on-error limit ({MAXIMUM_ERROR_RETRIES}).")
                    break
                continue
            if existing_status is not None:
                # Check if new timestamp is later than existing timestamp
                existing_deployment = existing_status['activeDeployment'] if 'activeDeployment' in existing_status else existing_status['failedDeployment']
                if status['timestamp'] <= existing_deployment['timestamp']:
                    # Do not parse old version
                    continue
            status_str = status['status']
            if status_str not in BOT_DEPLOYMENT_PROCESS:
                raise ValueError(f"Unknown status: {status_str}.")
            new_current_deployment_process = BOT_DEPLOYMENT_PROCESS.index(status_str)
            if new_current_deployment_process != current_deployment_process:
                # Completed new step(s)
                for step in BOT_DEPLOYMENT_PROCESS[:new_current_deployment_process + 1]:
                    if step in completed_processes:
                        continue
                    spinner.succeed(step)
                    completed_processes.append(step)
                current_deployment_process = new_current_deployment_process
                if current_deployment_process + 1 < len(BOT_DEPLOYMENT_PROCESS):
                    # If next step exists, set spinner to next step
                    spinner.start(
                        f"Waiting for next step: {BOT_DEPLOYMENT_PROCESS[current_deployment_process + 1]}"
                    )
                else:
                    # Next step does not exist; wait for final deployment confirmation
                    spinner.start("Waiting for active_deployment confirmation...")
            if 'activeDeployment' in status:
                new_active_deployment = status['activeDeployment']
                if existing_status is None \
                        or ('activeDeployment' in existing_status
                            and new_active_deployment['version'] >= existing_status['activeDeployment']['version']):
                    spinner.succeed("active_deployment")
                    print(f"New active deployment: {new_active_deployment}")
                    break
            if 'failedDeployment' in status:
                new_failed_deployment = status['failedDeployment']
                if existing_status is None \
                        or ('failedDeployment' in existing_status
                            and new_failed_deployment['version'] > existing_status['failedDeployment']['version']):
                    spinner.fail("failed_deployment")
                    display_logs(
                        get_logs(
                            bot_uid=bot_uid,
                            errors=True
                        )
                    )
                    print("Full logs can be checked with the display_logs and get_logs functions.")

                    break
            elapsed_time = time.time() - start_time
            if elapsed_time > 2 * 60:
                print("This deployment is taking an unexpectedly long time.")
            time.sleep(sleep)


class Deployment(TypedDict):
    timestamp: int
    version: int


class BotStatus(TypedDict):
    status: str
    timestamp: int
    activeDeployment: Optional[Deployment]
    failedDeployment: Optional[Deployment]


def get_bot_status(bot_uid: str) -> BotStatus:
    """Gets the status of the bot.

    :param bot_uid:
    :return:
    """
    auth = get_auth()
    try:
        req = requests.get(
            url=ENDPOINT,
            params={
                'uid': auth.uid,
                'key': auth.key,
                'bot_uid': bot_uid,
                'item': 'status'
                }
        )
        try:
            req.raise_for_status()
        except Exception:
            raise RuntimeError(req.json())
    except Exception:
        raise RuntimeError(f"Failed to retrieve status for bot {bot_uid}.")
    return req.json()


def share_bot(bot_uid: str) -> str:
    """Displays the url, a QR code, along with additional guidance.

    :param bot_uid:
    :return: The url for the bot.
    """
    url = f"chai://chai.ml/{bot_uid}"

    qr_code = segno.make_qr(url)
    print("Scan the QR code with your phone to start a chat in the app!")
    print(f"Or check it out at {url}")

    if IS_NOTEBOOK:
        show_qr(qr_code)
    else:
        qr_code.terminal()

    def save():
        path = input("Save QR code to: (Press [Enter] for default: 'qr.png') ")
        if len(path) == 0:
            path = "qr.png"
        qr_code.save(path, scale=10)
        print(f"Saved QR code to {path}.")

    def open_():
        qr_code.show(scale=10)

    actions = {
        "s": (save, "Save this QR code to an image file.")
    }
    if not IS_NOTEBOOK:
        actions["o"] = (open_, "Open QR code in external viewer.")

    while True:
        print("\nEnter one of the following keys to perform additional actions (or [Enter] to exit):")
        for key, action in actions.items():
            print(f" [{key}] {action[1]}")

        input_key = input().lower()
        if input_key in actions:
            actions[input_key][0]()
        else:
            print("Exiting.")
            break

    return url


def delete_bot(bot_uid: str) -> str:
    """
    Uses an HTTPS request to trigger deletion of bot with specified UID.

    :param bot_uid:
    :return: The url for the bot.
    """
    auth = get_auth()
    try:
        req = requests.delete(
            url=ENDPOINT,
            params={'uid': auth.uid, 'key': auth.key, 'bot_uid': bot_uid}
        )
        try:
            req.raise_for_status()
        except Exception:
            raise RuntimeError(req.json())
    except Exception as exc:
        raise RuntimeError(f"Failed to delete bot {bot_uid}.")

    print(f"Successfully deleted {bot_uid}.")
    return bot_uid

advertise_deployed_bot = share_bot
